import json
import sklearn.metrics
import torch
import torch.nn as nn
import torch.nn.functional as F

class DenseLogisticRegression(nn.Module):
    def __init__(self, input_dim):
        super().__init__()
        self.fc1 = nn.Linear(input_dim, 50)
        self.relu1 = nn.ReLU()
        self.fc2 = nn.Linear(50, 20)
        self.relu2 = nn.ReLU()
        self.fc3 = nn.Linear(20, 1)        
        self.init_weights()

    def init_weights(self):
        initrange = 0.5
        self.fc1.weight.data.uniform_(-initrange, initrange)
        self.fc1.bias.data.zero_()
        self.fc2.weight.data.uniform_(-initrange, initrange)
        self.fc2.bias.data.zero_()
        self.fc3.weight.data.uniform_(-initrange, initrange)
        self.fc3.bias.data.zero_()

    def forward(self, inputs):
        fc1 = self.fc1(inputs)
        fc2 = self.fc2(self.relu1(fc1))
        fc3 = self.fc3(self.relu2(fc2))
        return torch.sigmoid(fc3)

    def get_metrics(self, y, pred):
        auc_roc = sklearn.metrics.roc_auc_score(y, pred)
        auc_pr = sklearn.metrics.average_precision_score(y, pred)
        return {
            'auc_roc': auc_roc,
            'auc_pr': auc_pr
        }

    def optimizer(self):
        return torch.optim.SGD(self.parameters(), lr=1.0)

    @classmethod
    def criterion(cls):
        return torch.nn.BCELoss(size_average=True)


class DenseRegression(nn.Module):
    def __init__(self, input_dim, exp=False):
        super().__init__()
        self.exp = exp
        self.fc1 = nn.Linear(input_dim, 1)
        self.init_weights()

    def init_weights(self):
        initrange = 0.5
        self.fc1.weight.data.uniform_(-initrange, initrange)
        self.fc1.bias.data.zero_()

    def forward(self, inputs):
        fc1 = self.fc1(inputs)
        if self.exp:
            return torch.exp(fc1)
        else:
            return fc1

    def get_metrics(self, y, pred):
        r2 = sklearn.metrics.r2_score(y, pred)
        return {
            'r2': r2,
        }

    def optimizer(self):
        return torch.optim.SGD(self.parameters(), lr=50.0)

    @classmethod
    def criterion(cls):
        return torch.nn.L1Loss(reduction='sum')




import json
import sklearn.metrics
import time
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.utils.data import DataLoader


def generate_batch(batch):
    feature = torch.tensor([entry[0] for entry in batch], dtype=torch.float32)
    label = torch.tensor([entry[1] for entry in batch], dtype=torch.float32)
    label = label.unsqueeze(1)
    return feature, label


class Trainer():
    def __init__(self, model, train_data, dev_data):
        self.train_data = train_data
        self.dev_data = dev_data
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = model.to(self.device)
        self.criterion = model.criterion()
        self.optimizer = model.optimizer()
        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, 1, gamma=0.9)
        self.debug = False

    def train(self, batch_size=128):
        # Train the model
        train_loss = 0
        data = DataLoader(self.train_data, batch_size=batch_size, shuffle=True, collate_fn=generate_batch)
        y = []
        pred = []
        for i, (inputs, cls) in enumerate(data):
            y.extend(cls.tolist())
            self.optimizer.zero_grad()
            inputs, cls = inputs.to(self.device), cls.to(self.device)
            output = self.model(inputs)
            self.loss = self.criterion(output, cls)
            self.loss.backward()
            self.optimizer.step()

            pred.extend(output.cpu().tolist())
            train_loss += self.loss.item()

        # Adjust the learning rate
        self.scheduler.step()

        train_len = len(self.train_data)
        metrics = {
            'loss': train_loss / train_len
        }
        for k, v in self.model.get_metrics(y, pred).items():
            metrics[k] = v
        return metrics, y, pred
    
    def dev(self, batch_size=128):
        loss = 0
        data = DataLoader(self.dev_data, batch_size=batch_size, collate_fn=generate_batch)
        y = []
        pred = []
        for inputs, cls in data:
            y.extend(cls.tolist())
            inputs, cls = inputs.to(self.device), cls.to(self.device)

            with torch.no_grad():
                output = self.model(inputs)
                loss = self.criterion(output, cls)
                loss += loss.item()
                pred.extend(output.cpu().tolist())

        dev_len = len(self.dev_data)
        metrics = {
            'loss': loss / dev_len
        }
        for k, v in self.model.get_metrics(y, pred).items():
            metrics[k] = v
        return metrics, y, pred

    def predict(self, infer_data, batch_size=128):
        data = DataLoader(infer_data, batch_size=batch_size, collate_fn=generate_batch)
        pred = []
        for inputs, _ in data:
            inputs = inputs.to(self.device)
            with torch.no_grad():
                output = self.model(inputs)
                pred.extend(output.cpu().tolist())

        return pred

    def train_loop(self, batch_size=128, n_epochs=50):
        min_dev_loss = float('inf')
        min_dev_metrics = None
        all_dev_metrics = []
        for epoch in range(n_epochs):
            start_time = time.time()
            train_metric, train_y, train_pred = self.train(batch_size)
            dev_metric, dev_y, dev_pred = self.dev(batch_size)
            secs = int(time.time() - start_time)
            mins = secs / 60
            secs = secs % 60

            print('Epoch: %d, | time in {%d} minutes, {%d} seconds.' % (epoch + 1, mins, secs))
            print('Train loss: %.4f' % train_metric['loss'])
            print('Dev loss: %.4f' % dev_metric['loss'])
            dev_metric['epoch'] = epoch + 1
            all_dev_metrics.append(dev_metric)
            dev_loss = dev_metric['loss']
            if dev_loss < min_dev_loss:
                min_dev_loss = dev_loss
                min_dev_metrics = dev_metric
            
            for k, v in dev_metric.items():
                print('Dev %s: %.4f' % (k, v))

            if self.debug:
                for (y, pred) in zip(train_y[:10], train_pred[:10]):
                    print(y, pred)
                
                for (y, pred) in zip(dev_y[:10], dev_pred[:10]):
                    print(y, pred)

        return all_dev_metrics, min_dev_metrics
